#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2009 John Beard john.j.beard@gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This extension renders a wireframe sphere constructed from lines of latitude
and lines of longitude.

The number of lines of latitude and longitude is independently variable. Lines
of latitude and longtude are in separate subgroups. The whole figure is also in
its own group.

The whole sphere can be tilted towards or away from the veiwer by a given
number of degrees. If the whole sphere is then rotated normally in Inkscape,
any position can be achieved.

There is an option to hide the lines at the back of the sphere, as if the
sphere were opaque.
"""
# FIXME: Lines of latitude only have an approximation of the function needed
#            to hide the back portion. If you can derive the proper equation,
#            please add it in.
#            Line of longitude have the exact method already.
#            Workaround: Use the Inkscape ellipse tool to edit the start and end
#            points of the lines of latitude to end at the horizon circle.
#
# TODO:  Add support for odd numbers of lines of longitude. This means breaking
#        the line at the poles, and having two half ellipses for each line.
#        The angles at which the ellipse arcs pass the poles are not constant and
#        need to be derived before this can be implemented.
# TODO:  Add support for prolate and oblate spheroids
#
#    0.10    2009-10-25  First version. Basic spheres supported.
#                        Hidden lines of latitude still not properly calculated.
#                        Prolate and oblate spheroids not considered.

from math import acos, atan, cos, pi, sin, tan

import inkex
from inkex.localization import inkex_gettext as _

# add a tiny value to the ellipse radii, so that if we get a
# zero radius, the ellipse still shows up as a line
EPSILON = 0.001


class WireframeSphere(inkex.GenerateExtension):
    """Writeframe extension, generate a wireframe"""

    container_label = "WireframeSphere"

    def container_transform(self):
        transform = super(WireframeSphere, self).container_transform()
        if self.options.TILT < 0:
            transform *= inkex.Transform(scale=(1, -1))
        return transform

    def add_arguments(self, pars):
        pars.add_argument("--num_lat", type=int, dest="NUM_LAT", default=19)
        pars.add_argument("--num_long", type=int, dest="NUM_LONG", default=24)
        pars.add_argument("--radius", type=float, dest="RADIUS", default=100.0)
        pars.add_argument("--tilt", type=float, dest="TILT", default=35.0)
        pars.add_argument("--rotation", type=float, dest="ROT_OFFSET", default=4)
        pars.add_argument(
            "--hide_back", type=inkex.Boolean, dest="HIDE_BACK", default=False
        )

    def generate(self):
        opt = self.options

        # PARAMETER PROCESSING
        if opt.NUM_LONG % 2 != 0:  # lines of longitude are odd : abort
            inkex.errormsg(_("Please enter an even number of lines of longitude."))
            return

        radius = self.svg.unittouu(str(opt.RADIUS) + "px")
        tilt = abs(opt.TILT) * (pi / 180)  # Convert to radians
        rotate = opt.ROT_OFFSET * pi / 180  # Convert to radians

        # only process longitudes if we actually want some
        if opt.NUM_LONG > 0:
            # Yieled elements are added to generated container
            yield self.longitude_lines(opt.NUM_LONG, tilt, radius, rotate)

        if opt.NUM_LAT > 0:
            # Yieled elements are added to generated container
            # Account for the fact that we loop over N-1 elements
            yield self.latitude_lines(opt.NUM_LAT + 1, tilt, radius)

        # THE HORIZON CIRCLE - circle, centred on the sphere centre
        yield self.draw_ellipse((radius, radius), (0, 0))

    def longitude_lines(self, number, tilt, radius, rotate):
        """Add lines of latitude as a group"""
        # GROUP FOR THE LINES OF LONGITUDE
        grp_long = inkex.Group()
        grp_long.set("inkscape:label", "Lines of Longitude")

        # angle between neighbouring lines of longitude in degrees
        # delta_long = 360.0 / number

        for i in range(0, number // 2):
            # The longitude of this particular line in radians
            long_angle = rotate + (i * (360.0 / number)) * (pi / 180.0)
            if long_angle > pi:
                long_angle -= 2 * pi
            # the rise is scaled by the sine of the tilt
            # length     = sqrt(width*width+height*height)  #by pythagorean theorem
            # inverse    = sin(acos(length/so.RADIUS))
            inverse = abs(sin(long_angle)) * cos(tilt)

            rads = (radius * inverse + EPSILON, radius)

            # The rotation of the ellipse to get it to pass through the pole (degs)
            rotation = atan(
                (radius * sin(long_angle) * sin(tilt)) / (radius * cos(long_angle))
            ) * (180.0 / pi)

            # remove the hidden side of the ellipses if required
            # this is always exactly half the ellipse, but we need to find out which half
            start_end = (0, 2 * pi)  # Default start and end angles -> full ellipse
            if self.options.HIDE_BACK:
                if long_angle <= pi / 2:  # cut out the half ellispse that is hidden
                    start_end = (pi / 2, 3 * pi / 2)
                else:
                    start_end = (3 * pi / 2, pi / 2)

            # finally, draw the line of longitude
            # the centre is always at the centre of the sphere
            elem = grp_long.add(self.draw_ellipse(rads, (0, 0), start_end))
            # the rotation will be applied about the group centre (the centre of the sphere)
            elem.transform = inkex.Transform(rotate=(rotation,))
        return grp_long

    def latitude_lines(self, number, tilt, radius):
        """Add lines of latitude as a group"""
        # GROUP FOR THE LINES OF LATITUDE
        grp_lat = inkex.Group()
        grp_lat.set("inkscape:label", "Lines of Latitude")

        # Angle between the line of latitude (subtended at the centre)
        delta_lat = 180.0 / number

        for i in range(1, number):
            # The angle of this line of latitude (from a pole)
            lat_angle = (delta_lat * i) * (pi / 180)

            # The width of the LoLat (no change due to projection)
            # The projected height of the line of latitude
            rads = (
                radius * sin(lat_angle),  # major
                (radius * sin(lat_angle) * sin(tilt)) + EPSILON,  # minor
            )

            # The x position is the sphere center, The projected y position of the LoLat
            pos = (0, radius * cos(lat_angle) * cos(tilt))

            if self.options.HIDE_BACK:
                if lat_angle > tilt:  # this LoLat is partially or fully visible
                    if lat_angle > pi - tilt:  # this LoLat is fully visible
                        grp_lat.add(self.draw_ellipse(rads, pos))
                    else:  # this LoLat is partially visible
                        proportion = (
                            -(acos(tan(lat_angle - pi / 2) / tan(pi / 2 - tilt))) / pi
                            + 1
                        )
                        # make the start and end angles (mirror image around pi/2)
                        start_end = (pi / 2 - proportion * pi, pi / 2 + proportion * pi)
                        grp_lat.add(self.draw_ellipse(rads, pos, start_end))

            else:  # just draw the full lines of latitude
                grp_lat.add(self.draw_ellipse(rads, pos))
        return grp_lat

    def draw_ellipse(self, r_xy, c_xy, start_end=(0, 2 * pi)):
        """Creates an elipse with all the required sodipodi attributes"""
        path = inkex.PathElement.arc(
            c_xy,
            *r_xy,
            start=start_end[0],
            end=start_end[1],
            open="true",
            arctype="arc",
        )
        path.style = {
            "stroke": "#000000",
            "stroke-width": str(self.svg.unittouu("1px")),
            "fill": "none",
        }
        return path


if __name__ == "__main__":
    WireframeSphere().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2006 Aaron Spike, aaron@ekips.org
# Copyright (C) 2010-2012 Nicolas Dufour, nicoduf@yahoo.fr
# (Windows support and various fixes)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Export to Gimp's XCF file format including Grids and Guides.
"""

import os
from collections import OrderedDict

import inkex
from inkex.base import TempDirMixin
from inkex.command import take_snapshot, call
from inkex.localization import inkex_gettext as _


class GimpXcf(inkex.OutputExtension):
    """
    Provide a quick and dirty way of using gimp to output an xcf from Inkscape.

    Both Inkscape and Gimp must be installed for this extension to work.
    """

    dir_prefix = "gimp-out-"

    def add_arguments(self, pars):
        pars.add_argument("--tab", dest="tab")
        pars.add_argument(
            "-d", "--guides", type=inkex.Boolean, help="Save the Guides in the XCF"
        )
        pars.add_argument(
            "-r", "--grid", type=inkex.Boolean, help="Save the Grid with the .XCF"
        )
        pars.add_argument(
            "-b", "--background", type=inkex.Boolean, help="Add background color"
        )
        pars.add_argument(
            "-i", "--dpi", type=float, default=96.0, help="File resolution"
        )

    def get_guides(self):
        """Generate a list of horzontal and vertical only guides"""
        horz_guides = []
        vert_guides = []
        # Grab all guide tags in the namedview tag
        # TODO: compute intersection of guide with page instead of discarding those
        # whose control points are outside of the page
        for guide in self.svg.namedview.get_guides():
            if guide.is_horizontal:
                # GIMP doesn't like guides that are outside of the image
                if 0 < guide.raw_position.y < self.svg.viewbox_height:
                    # The origin is at the top in GIMP land # TODO check this (maybe use position instead of raw_position?)
                    horz_guides.append(str(guide.raw_position.y))
            elif guide.is_vertical:
                # GIMP doesn't like guides that are outside of the image
                if 0 < guide.raw_position.x < self.svg.viewbox_width:
                    vert_guides.append(str(guide.raw_position.x))

        return ("h", " ".join(horz_guides)), ("v", " ".join(vert_guides))

    def get_grid(self):
        """Get the grid if asked for and return as gimpfu script"""
        scale = (self.svg.scale) * (self.options.dpi / 96.0)
        # GIMP only allows one rectangular grid
        xpath = "sodipodi:namedview/inkscape:grid[@type='xygrid' and (not(@units) or @units='px')]"
        if self.svg.xpath(xpath):
            node = self.svg.getElement(xpath)
            for attr, default, target in (
                ("spacing", 1, "spacing"),
                ("origin", 0, "offset"),
            ):
                fmt = {"target": target}
                for dim in "xy":
                    # These attributes could be nonexistent
                    unit = float(node.get(attr + dim, default))
                    unit = self.svg.uutounit(unit, "px") * scale
                    fmt[dim] = int(round(float(unit)))
                yield "(gimp-image-grid-set-{target} img {x} {y})".format(**fmt)

    @property
    def docname(self):
        """Get the document name suitable for export"""
        return self.svg.get("sodipodi:docname") or "document"

    def save(self, stream):
        pngs = OrderedDict()
        valid = False

        for node in self.svg.xpath("/svg:svg/*[name()='g' or @style][@id]"):
            if not len(node):  # pylint: disable=len-as-condition
                # Ignore empty layers
                continue

            valid = True
            node_id = node.get("id")
            name = node.get("inkscape:label", node_id)

            pngs[name] = take_snapshot(
                self.document,
                dirname=self.tempdir,
                name=name,
                dpi=int(self.options.dpi),
                export_id=node_id,
                export_id_only=True,
                export_area_page=True,
                export_background_opacity=int(bool(self.options.background)),
            )

        if not valid:
            inkex.errormsg(_("This extension requires at least one non empty layer."))
            return

        xcf = os.path.join(self.tempdir, "{}.xcf".format(self.docname))
        script_fu = """
(tracing 1)
(define
  (png-to-layer img png_filename layer_name)
  (let*
    (
      (png (car (file-png-load RUN-NONINTERACTIVE png_filename png_filename)))
      (png_layer (car (gimp-image-get-active-layer png)))
      (xcf_layer (car (gimp-layer-new-from-drawable png_layer img)))
    )
    (gimp-image-add-layer img xcf_layer -1)
    (gimp-drawable-set-name xcf_layer layer_name)
  )
)
(let*
  (
    (img (car (gimp-image-new 200 200 RGB)))
  )
  (gimp-image-set-resolution img {dpi} {dpi})
  (gimp-image-undo-disable img)
  (for-each
    (lambda (names)
      (png-to-layer img (car names) (cdr names))
    )
    (map cons '("{files}") '("{names}"))
  )

  (gimp-image-resize-to-layers img)
""".format(
            dpi=self.options.dpi,
            files='" "'.join(pngs.values()),
            names='" "'.join(list(pngs)),
        )

        if self.options.guides:
            for dim, guides in self.get_guides():
                script_fu += """
  (for-each
    (lambda ({d}Guide)
      (gimp-image-add-{d}guide img {d}Guide)
    )
    '({g})
  )""".format(d=dim, g=guides)

        # Grid
        if self.options.grid:
            for fu_let in self.get_grid():
                script_fu += "\n" + fu_let + "\n"

        script_fu += """
  (gimp-image-undo-enable img)
  (gimp-file-save RUN-NONINTERACTIVE img (car (gimp-image-get-active-layer img)) "{xcf}" "{xcf}"))
(gimp-quit 0)
            """.format(xcf=xcf)

        call(
            "gimp",
            "-b",
            "-",
            i=True,
            batch_interpreter="plug-in-script-fu-eval",
            stdin=script_fu,
        )

        with open(xcf, "rb") as fhl:
            stream.write(fhl.read())


if __name__ == "__main__":
    GimpXcf().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2022 Jonathan Neuhauser (jonathan.neuhauser@outlook.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Parser for HP/GL2 Documents (includes ordinary HPGL)"""

import pyparsing as pp

from inkex.base import SvgOutputMixin
import inkex
from inkex.localization import inkex_gettext as _


from hpgl_input_sm import HPGLStateMachine
from hpgl_parser import build_parser


class Hpgl2Input(inkex.InputExtension):
    """InputExtension to import HP/GL2 documents."""

    def add_arguments(self, pars):
        pars.add_argument(
            "--width", type=float, default=8, help="Width of the picture frame [in]"
        )
        pars.add_argument(
            "--height", type=float, default=10, help="Height of the picture frame [in]"
        )
        pars.add_argument(
            "--resolution", type=float, default=1016.0, help="Resolution (plu/in)"
        )
        pars.add_argument(
            "--bake-transforms",
            type=inkex.Boolean,
            default=True,
            help=(
                "Bake transforms (disabling is helpful for debugging IP/IR/SC commands)"
            ),
        )
        pars.add_argument(
            "--break-apart",
            type=inkex.Boolean,
            default=True,
            help="Break apart subpaths",
        )

    def load(self, stream):
        """Load the document, making sure that we operate on a decoded string"""
        res = stream.read()
        if isinstance(res, str):
            return res
        return res.decode()

    def effect(self):
        # interpret HPGL data
        doc: inkex.SvgDocumentElement = SvgOutputMixin.get_template(
            width=self.options.width * 25.4,
            height=self.options.height * 25.4,
            unit="mm",
        )

        layer = doc.getroot().add(inkex.Layer())

        plu_to_svg = inkex.Transform(
            scale=(
                0.025 / self.options.resolution * 1016,
                -0.025 / self.options.resolution * 1016,
            ),
            translate=(0, self.options.height * 25.4),
        )
        layer.transform = plu_to_svg

        document_parser = build_parser(HPGLStateMachine(layer, self.options))
        try:
            result = document_parser.parse_string(self.document, parse_all=True)
            # bake in the transforms
            if self.options.bake_transforms:
                layer.bake_transforms_recursively()
            unknown = result.get("u")
            if unknown:
                inkex.errormsg(
                    _(
                        "Unsupported commands encountered. "
                        "The following commands were ignored:"
                    )
                )
                inkex.errormsg(unknown)

        except pp.ParseException as parse_err:
            inkex.errormsg(parse_err.explain())

        # deliver document to inkscape
        self.document = doc


if __name__ == "__main__":
    Hpgl2Input().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2011 Felipe Correa da Silva Sanches <juca@members.fsf.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Extension for converting svg fonts to layers"""

import inkex


class SvgFontToLayers(inkex.EffectExtension):
    """Convert an svg font to layers"""

    def add_arguments(self, pars):
        pars.add_argument(
            "--count",
            type=int,
            default=30,
            help="Stop making layers after this number of glyphs.",
        )

    def flip_cordinate_system(self, elem, emsize, baseline):
        """Scale and translate the element's path, returns the path object"""
        path = elem.path
        path.scale(1, -1, inplace=True)
        path.translate(0, int(emsize) - int(baseline), inplace=True)
        return path

    def effect(self):
        # TODO: detect files with multiple svg fonts declared.
        # Current code only reads the first svgfont instance
        font = self.svg.defs.findone("svg:font")
        if font is None:
            return inkex.errormsg("There are no svg fonts")
        # setwidth = font.get("horiz-adv-x")
        baseline = font.get("horiz-origin-y")
        if baseline is None:
            baseline = 0

        guidebase = self.svg.viewbox_height - baseline

        fontface = font.findone("svg:font-face")

        # TODO: where should we save the font family name?
        # fontfamily = fontface.get("font-family")
        emsize = fontface.get("units-per-em")

        # TODO: should we guarantee that <svg:font horiz-adv-x> equals <svg:font-face units-per-em> ?
        caps = int(fontface.get("cap-height", 0))
        xheight = int(fontface.get("x-height", 0))
        ascender = int(fontface.get("ascent", 0))
        descender = int(fontface.get("descent", 0))

        self.svg.set("width", emsize)
        self.svg.namedview.add_guide(guidebase, True, "baseline")
        self.svg.namedview.add_guide(guidebase - ascender, True, "ascender")
        self.svg.namedview.add_guide(guidebase - caps, True, "caps")
        self.svg.namedview.add_guide(guidebase - xheight, True, "xheight")
        self.svg.namedview.add_guide(guidebase + descender, True, "decender")

        # TODO: missing-glyph
        count = 0
        for glyph in font.findall("svg:glyph"):
            unicode_char = glyph.get("unicode")
            if unicode_char is None:
                continue

            layer = self.svg.add(inkex.Layer.new("GlyphLayer-" + unicode_char))
            # glyph layers (except the first one) are innitially hidden
            if count != 0:
                layer.style["display"] = "none"

            ############################
            # Option 1:
            # Using clone (svg:use) as childnode of svg:glyph

            # use = self.get_or_create(glyph, inkex.Use())
            # use.href = group
            # TODO: This code creates <use> nodes but they do not render on svg fonts dialog. why?

            ############################
            # Option 2:
            # Using svg:paths as childnodes of svg:glyph
            for elem in glyph.findall("svg:path"):
                new_path = layer.add(inkex.PathElement())
                new_path.path = self.flip_cordinate_system(elem, emsize, baseline)

            ############################
            # Option 3:
            # Using curve description in d attribute of svg:glyph
            path = layer.add(inkex.PathElement())
            path.path = self.flip_cordinate_system(glyph, emsize, baseline)

            count += 1
            if count >= self.options.count:
                break


if __name__ == "__main__":
    SvgFontToLayers().run()



#!/usr/bin/env python3
"""Replace color extension"""

import inkex


class ReplaceColor(inkex.ColorExtension):
    """Replace color in SVG with another"""

    pass_rgba = True

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument(
            "-f",
            "--from_color",
            default=inkex.Color("black"),
            type=inkex.Color,
            help="Replace color",
        )
        pars.add_argument(
            "-t",
            "--to_color",
            default=inkex.Color("red"),
            type=inkex.Color,
            help="By color",
        )
        pars.add_argument(
            "-i",
            "--ignore_opacity",
            default=True,
            type=inkex.Boolean,
            help="Whether color should be replaced regardless of opacity match",
        )

    def modify_color(self, name, color):  # color is rgba
        if self.options.from_color.to_rgb() == color.to_rgb() and (
            self.options.ignore_opacity
            or abs(self.options.from_color.to_rgba().alpha - color.alpha) < 0.01
        ):
            return self.options.to_color.to_rgba()
        return color


if __name__ == "__main__":
    ReplaceColor().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2020-2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""SVG -> XAML Base converter class, handling most styles"""
from typing import Optional, Union

import inkex
from inkex.transforms import Transform

from .config import GLOBAL_SETTINGS
from ..xamlobjects import (
    XAMLObject,
    BaseElement,
    Shape,
    GeometryGroup,
    Geometry,
    BlurEffect,
    Brush,
    VisualBrush,
    DrawingBrush,
)
from .utils import convert_attributes


class BaseParser:
    """Base class for parsers from SVG objects to XAML"""

    def __init__(self, svgelement: inkex.BaseElement):
        self.element = svgelement

    def to_xaml_objects(self):
        """Base method to convert self.element to an XAML object for a Drawing"""

    def to_canvas_objects(self):
        """Base method to convert self.element to an XAML object for a Canvas"""

    def get_clip(
        self, xaml_parent: BaseElement, value=None
    ) -> Optional[Union[GeometryGroup, Geometry, str]]:
        """Converts a Clip to XAML (Geometry) representation. If value is not given,
        it is pulled from self.element. xaml_parent is used to evaluate transforms
        correctly"""
        if isinstance(xaml_parent, GeometryGroup):
            return None  # XAML doesn't support nested clips
        if value is None:
            if isinstance(self.element, inkex.ShapeElement):
                value = self.element.clip
        # Construct the child parser
        if isinstance(value, inkex.ClipPath):
            # ClipPaths in XAML are specified in the coordinate system of the clip
            # definition, while in SVG they are specified in userSpaceOnUse. So we have
            # to apply the inverse transform of the self.element to a copy of the
            # clippath. Most probably, the clippath is only referenced by one element.
            # If it is referenced multiple times, the clippath will be duplicated.
            transformcopy = value.transform
            # For elements which specify a Canvas coordinate,
            # the clip is relative to that coordinate
            left = xaml_parent.get("Canvas.Left", 0)
            top = xaml_parent.get("Canvas.Top", 0)
            additional_transform = inkex.Transform(translate=(left, top))
            # do we need to consider parent transforms?
            value.transform = value.transform @ (-additional_transform)
            from .structural import ClipPathParser

            parser = ClipPathParser(value)
            result: Union[GeometryGroup, Geometry, str, None] = parser.to_clippath()
            if result is not None and isinstance(result, GeometryGroup):
                result = parser.simplify_clip(result)
            value.transform = transformcopy
            return result
        return None

    def get_mask(self, xaml_parent, value=None):
        """Converts a Mask to XAML (Geometry) representation. If value is not given,
        it is pulled from self.element. xaml_parent is used to evaluate transforms
        correctly"""
        if value is None:
            if isinstance(self.element, inkex.ShapeElement):
                value = self.element.specified_style()("mask")
        if value is None:
            return None
        return self.get_mask_pattern(xaml_parent, value)

    def get_blur(self) -> Optional[BlurEffect]:
        """Checks if an object has a Blur filter applied, and in that case returns the
        equivalent XAML object"""
        value = self.element.specified_style()("filter")
        # Only if it's a single filter with a single blur primitive
        if (
            value is None
            or len(value) != 1
            or len(value[0]) != 1
            or value[0][0].TAG != "feGaussianBlur"
        ):
            return None
        result = BlurEffect()
        std_dev = value[0][0].get("stdDeviation")
        std_dev = float(std_dev.split()[0]) * 3.1  # empirical factor
        if std_dev < 0.5:
            # Ignore
            return None
        if 0.5 <= std_dev < 1:
            std_dev = 1
        result.add_property("Radius", std_dev)
        return result

    def set_mask_pattern_viewbox(self, value, pattern, parsed):
        """Set the viewbox of a mask or pattern."""
        mask = isinstance(value, inkex.Mask)

        viewboxprops = [value.x, value.y, value.width, value.height]
        units = value.patternUnits if not mask else value.maskUnits
        if all(x == 0 for x in viewboxprops) and mask:
            # Fall back to absolute units.
            bbx = value.shape_box()
            if bbx is not None:
                viewboxprops = [
                    bbx.x.minimum * 0.9,
                    bbx.y.minimum * 0.9,
                    bbx.width * 1.2,
                    bbx.height * 1.2,
                ]
            units = "userSpaceOnUse"

        if GLOBAL_SETTINGS.avalonia:
            viewboxprops = [float(i) for i in viewboxprops]
            pattern.add_properties(
                {
                    "Width": viewboxprops[0] + float(viewboxprops[2]),
                    "Height": viewboxprops[1] + float(viewboxprops[3]),
                }
            )
            parsed.add_properties(
                {
                    "SourceRect": " ".join(map(str, viewboxprops)),
                    "DestinationRect": " ".join(
                        map(
                            str, [-viewboxprops[0], -viewboxprops[1]] + viewboxprops[2:]
                        )
                    ),
                }
            )
        else:
            # ideally properly inherit this attribute, see pservers-pattern-04
            convert_attributes(
                None,
                parsed,
                [
                    [
                        "patternUnits",
                        "ViewportUnits",
                        {
                            "objectBoundingBox": "RelativeToBoundingBox",
                            "userSpaceOnUse": "Absolute",
                        },
                        "RelativeToBoundingBox",
                        units,
                    ],
                    [
                        "patternUnits",
                        "ViewboxUnits",
                        {
                            "objectBoundingBox": "RelativeToBoundingBox",
                            "userSpaceOnUse": "Absolute",
                        },
                        "RelativeToBoundingBox",
                        units,
                    ],
                ],
            )
            if not mask:
                parsed.add_properties(
                    {
                        "ViewboxUnits": "Absolute",
                    }
                )
            parsed.add_properties(
                {
                    "Viewbox": " ".join(map(str, viewboxprops)),
                    "Viewport": " ".join(map(str, viewboxprops)),
                }
            )
        return viewboxprops

    def set_pattern_transform(self, pattern, parsed, value, viewboxprops):
        """Set the transform on a pattern"""
        parsed.add_properties({"TileMode": "Tile"})
        if (value.get_viewbox() is not None) or not all(
            map(lambda x: x == "0", viewboxprops)
        ):
            # in this case, viewport and viewbox both have to be set together with
            # an initial transform of the pattern group
            vbox = value.get_viewbox()
            # pylint: disable=line-too-long
            if vbox is not None:
                # both are specified, apply viewbox as additional transform
                trans = inkex.Transform(
                    (
                        float(viewboxprops[2]) / vbox[2],
                        0,
                        0,
                        float(viewboxprops[2]) / vbox[3],
                        viewboxprops[0],
                        viewboxprops[1],
                    )
                )
            else:
                trans = inkex.Transform((1, 0, 0, 1, viewboxprops[0], viewboxprops[1]))

            if trans is not None:
                pattern.set_transform(trans)
        if not GLOBAL_SETTINGS.avalonia:
            parsed.set_transform(value.patternTransform)

    def get_mask_pattern(
        self, xaml_parent, value: Union[inkex.Mask, inkex.Pattern], opacity=1
    ):
        """Routine to convert Masks and Patterns to XAML"""
        lowlevel = True
        if GLOBAL_SETTINGS.avalonia:
            # Avalonia doesn't support DrawingBrush, so we have to
            # use a VisualBrush. Currently, pattern-transform is not supported.
            # In Avalonia 0.11, we'll also have a Transform attribute
            # (https://github.com/AvaloniaUI/Avalonia/pull/6344)
            lowlevel = False
        # in WPF, the ViewBox indicates which part of the VisualBrush we use
        # the ViewPort indicates the position of the first tile

        parsed: Brush = VisualBrush()
        if lowlevel:
            parsed = DrawingBrush()
        mask = isinstance(value, inkex.Mask)
        from .structural import GroupParser

        pps = GroupParser(value if mask else value.get_effective_parent())

        pattern: XAMLObject
        if lowlevel:
            pattern = pps.to_xaml_objects()
        else:
            pattern = pps.to_canvas_objects()

        viewboxprops = self.set_mask_pattern_viewbox(value, pattern, parsed)
        parsed.add_property("Drawing" if lowlevel else "Visual", pattern)
        if not mask:
            self.set_pattern_transform(pattern, parsed, value, viewboxprops)
        if isinstance(xaml_parent, Shape) and xaml_parent is not None:
            left = xaml_parent.get("Canvas.Left", None)
            top = xaml_parent.get("Canvas.Top", None)
            if left is not None and top is not None:
                # in a canvas, we have to apply a transform to the pattern
                pattern_transform = Transform(translate=(-float(left), -float(top))) @ (
                    inkex.Transform() if mask else value.patternTransform
                )
                parsed.set_transform(pattern_transform)
        pattern.add_property("Opacity", value.specified_style()("opacity") * opacity, 1)
        return parsed



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2020 Ellen Wasboe, ellen@wasbo.net
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
"""
Clean up path by removing duplicate nodes and interpolate close nodes.

Optionally:
    join start and end node of each subpath if distance < threshold
    join separate subpaths if end nodes closer than threshold
Joining subpaths can be done either by interpolating or straight line segment.
"""

import numpy as np

import inkex
from inkex import bezier
from inkex.paths import Move, Curve, Line, ZoneClose, Path
from inkex.localization import inkex_gettext as _


def get_length_bez_segment(prev_xy, bez):
    """Get bezier segment length.

    Parameters
    ----------
    prev_xy : list of float
        first node
        [x, y]
    bez : list of float
        command to second node
        [p2,p3,p4] as returned by Curve.to_bez()

    Returns
    -------
    length : float
    """
    # csp [p3, p0, p1]
    csp1 = [prev_xy, prev_xy, bez[0]]
    csp2 = [bez[1], bez[2], bez[2]]
    length = bezier.cspseglength(csp1, csp2)

    return length


def finish_subpath(path, prev_bez):
    """Finish subpath if stop before threshold met."""
    last_added = path[-1]
    if len(last_added.args) == 6:
        x2, y2, x3, y3, __, __ = last_added.args
        # pos of end node, ctrl points of last added
        path.append(Curve(x2, y2, x3, y3, prev_bez[2][0], prev_bez[2][1]))
    else:
        if last_added.letter == "M":
            path = Path()  # remove zero length path

    return path


def remove_last_subpath(path):
    """Remove last subpath from list of PathCommands."""
    while len(path) > 0:
        cmd = path.pop()
        if cmd.letter == "M":
            break
    return path


def get_distance_xy(xy1, xy2):
    """Calculate distance between coordiates [x1,y1] and [x2,y2]."""
    xdiff = xy2[0] - xy1[0]
    ydiff = xy2[1] - xy1[1]
    distance = np.sqrt(np.add(np.power(xdiff, 2), np.power(ydiff, 2)))

    return distance


class CleanUpPath(inkex.EffectExtension):
    def add_arguments(self, pars):
        pars.add_argument("--name", default="options")
        pars.add_argument("--minlength", type=float, default=0)
        pars.add_argument("--minUse", type=inkex.Boolean, default=False)
        pars.add_argument("--minlengthSub", type=float, default=0)
        pars.add_argument("--minUseSub", type=inkex.Boolean, default=False)
        pars.add_argument("--maxdistClose", type=float, default=0)
        pars.add_argument("--closeSub", type=inkex.Boolean, default=False)
        pars.add_argument("--maxdistJoin", type=float, default=0)
        pars.add_argument("--joinEndSub", type=inkex.Boolean, default=False)
        pars.add_argument("--allow_reverse", type=inkex.Boolean, default=True)
        pars.add_argument("--option_join", default="1")

    def effect(self):
        """Clean up path."""
        self.min_length = self.options.minlength
        self.min_length_sub = self.options.minlengthSub
        self.max_dist_close = self.options.maxdistClose
        self.max_dist_join = self.options.maxdistJoin
        if self.options.minUse is False:
            self.min_length = 0
        if self.options.minUseSub is False:
            self.min_length_sub = 0
        if self.options.closeSub is False:
            self.max_dist_close = -1
        if self.options.joinEndSub is False:
            self.max_dist_join = -1

        nInkEffect = 0
        for elem in self.svg.selection.filter_nonzero(inkex.PathElement):
            thisIsPath = True
            if elem.get("inkscape:path-effect") is not None:
                thisIsPath = False
                nInkEffect += 1

            if thisIsPath:
                self.path_d = elem.path.to_absolute()
                # split into dict of subpaths
                self.subpaths_dict = {
                    "subpaths": [sub for sub in self.path_d.subpath_iterator()]
                }
                self.remove_minlength_nodes()  # to curves and fill subpaths_dict
                if self.max_dist_close > -1:
                    self.close_subpaths()
                if self.max_dist_join > -1:
                    self.join_subpaths()

                # join dict of subpaths
                path = []
                for sub in self.subpaths_dict["subpaths"]:
                    path.extend(sub)
                elem.path = Path(path)

        if nInkEffect > 0:
            inkex.errormsg(
                _(
                    f"""{nInkEffect} selected elements have an
                inkscape:path-effect applied. These elements will be
                ignored to avoid confusing results. Apply Paths->Object
                to path (Shift+Ctrl+C) and retry .""",
                )
            )

    def remove_minlength_nodes(self):
        """Remove node if curve-length < threshold and fill subpaths_dict info."""
        threshold = self.min_length
        self.subpaths_dict["open"] = []
        self.subpaths_dict["xy_start"] = []
        self.subpaths_dict["xy_end"] = []

        pop_sub_dicts = []  # ids to remove due to shorter path than threshold
        for subno, sub in enumerate(self.subpaths_dict["subpaths"]):
            new_sub = Path()
            summed_length = 0
            total_sub_length = 0
            sum_next = False
            prev_bez = []
            for cmd in sub.proxy_iterator():
                if cmd.letter == "M":  # only first element
                    new_sub.append(Move(*cmd.args))
                elif cmd.letter == "Z":  # last element if closed path
                    if sum_next:  # last segment of subpath not added still
                        new_sub = finish_subpath(new_sub, prev_bez)
                    new_sub.append(ZoneClose(*cmd.args))
                else:
                    c = cmd.to_curve()
                    prev = cmd.previous_end_point
                    length = get_length_bez_segment([prev.x, prev.y], c.to_bez())
                    total_sub_length += length

                    if sum_next is False:  # fresh start, irrelevant history
                        if length > threshold:
                            new_sub.append(Curve(*c.args))  # add as is
                        else:
                            # keep including segments until total length >= threshold
                            summed_length = length
                            sum_next = True
                    else:
                        summed_length += length

                    if summed_length >= threshold and sum_next:
                        if new_sub[-1].letter == "M":
                            if threshold > 0:
                                new_sub.append(Curve(*c.args))  # add as is keep start
                            else:
                                pass  # ignore
                        else:
                            last_added = new_sub[-1]
                            # end of group of segments - correct last added node p4
                            # if sum > threshold avg pos from last+prev, add current
                            # if summ = threshold avg pos from last+current node
                            x2, y2, x3, y3, x4, y4 = last_added.args
                            if summed_length > threshold:
                                prev_end = cmd.previous_end_point
                                edit_x4 = 0.5 * (prev_end.x + x4)
                                edit_y4 = 0.5 * (prev_end.y + y4)
                            else:
                                __, __, __, __, x_4, y_4 = c.args
                                edit_x4 = 0.5 * (x_4 + x4)
                                edit_y4 = 0.5 * (y_4 + y4)
                            new_sub[-1] = Curve(x2, y2, x3, y3, edit_x4, edit_y4)
                            if summed_length > threshold:
                                new_sub.append(Curve(*c.args))

                        sum_next = False
                        summed_length = 0
                    prev_bez = c.to_bez()

            if sum_next:  # last segment of last subpath not added still
                new_sub = finish_subpath(new_sub, prev_bez)
            if total_sub_length > self.min_length_sub:
                self.subpaths_dict["subpaths"][subno] = new_sub
                if isinstance(new_sub[-1], ZoneClose):
                    self.subpaths_dict["open"].append(False)
                    last_args = new_sub[-2].args
                else:
                    self.subpaths_dict["open"].append(True)
                    last_args = new_sub[-1].args
                self.subpaths_dict["xy_start"].append(new_sub[0].args)
                self.subpaths_dict["xy_end"].append(last_args[-2:])
            else:
                pop_sub_dicts.append(subno)

        if len(pop_sub_dicts) > 0:
            pop_sub_dicts.sort(reverse=True)
            for no in pop_sub_dicts:
                self.subpaths_dict["subpaths"].pop(no)

    def close_subpaths(self):
        """Join end nodes of each subpath by interpolating coordinates."""
        for s, sub in enumerate(self.subpaths_dict["subpaths"]):
            if self.subpaths_dict["open"][s]:
                xy_start = self.subpaths_dict["xy_start"][s]
                xy_end = self.subpaths_dict["xy_end"][s]
                dist = get_distance_xy(xy_start, xy_end)
                if dist < self.max_dist_close:
                    avg_x = 0.5 * (xy_start[0] + xy_end[0])
                    avg_y = 0.5 * (xy_start[1] + xy_end[1])
                    sub[0] = Move(avg_x, avg_y)
                    end_cmd = sub[-1]
                    x2, y2, x3, y3, __, __ = end_cmd.args
                    sub[-1] = Curve(x2, y2, x3, y3, avg_x, avg_y)
                    sub.append(ZoneClose())
                    self.subpaths_dict["xy_start"][s] = [avg_x, avg_y]
                    self.subpaths_dict["xy_end"][s] = [avg_x, avg_y]
                    self.subpaths_dict["open"][s] = False
                    self.subpaths_dict["subpaths"][s] = sub

    def find_join_matches(self):
        """Find which subpath to join to which and to which end."""

        def find_id_min_dist(subno, join_to, xy_this, xy_all):
            id_min_dist = None
            min_distance = self.max_dist_join
            ids_test = np.where(join_to == -1)
            # avoid join to self
            id2test = np.delete(ids_test[0], np.where(ids_test[0] == subno))
            if id2test.size > 0:
                # calculate distances in x/y direction
                # find shortest distance if less than minimum
                xy_search = xy_all[id2test]
                xdiff = np.subtract(xy_search[:, 0], xy_this[0])
                ydiff = np.subtract(xy_search[:, 1], xy_this[1])
                distance = np.sqrt(np.add(np.power(xdiff, 2), np.power(ydiff, 2)))
                min_distance = np.nanmin(distance)
                if min_distance < self.max_dist_join:
                    id_min_list = np.where(distance == min_distance)
                    id_min = id_min_list[0]
                    id_min_dist = id2test[id_min[0]]

            return (id_min_dist, min_distance)

        def join_to(subno, j_to1, j_to2, xy1, xy2):
            id_join = None
            reverse = False
            if j_to1[subno] == -1:  # not joined yet
                xy_this = xy1[subno]
                id_join, min_distance = find_id_min_dist(subno, j_to2, xy_this, xy2)
                if self.options.allow_reverse:
                    id_join_r, min_dist_r = find_id_min_dist(subno, j_to1, xy_this, xy1)
                    if id_join_r is not None:
                        if id_join is not None:
                            if min_dist_r < min_distance:
                                reverse = True
                        else:
                            reverse = True
                    if reverse:
                        min_distance = min_dist_r
                        id_join = id_join_r

            return (id_join, reverse)

        def switch_start_end(xy_start, xy_end, id_join):
            x, y = xy_start[id_join]
            xy_start[id_join] = xy_end[id_join]
            xy_end[id_join] = [x, y]
            return (xy_start, xy_end)

        nSub = len(self.subpaths_dict["subpaths"])
        # initiate arrays
        j_end2 = np.full(nSub, -1)  # sub start index to join
        j_start2 = np.copy(j_end2)
        j_end2end = np.zeros(nSub, dtype=bool)
        j_start2start = np.zeros(nSub, dtype=bool)
        ids_closed = []
        xy_start = self.subpaths_dict["xy_start"]
        xy_end = self.subpaths_dict["xy_end"]
        for s in range(nSub):
            if self.subpaths_dict["open"][s] is False:
                ids_closed.append(s)
                xy_start[s] = [np.nan, np.nan]
                xy_end[s] = [np.nan, np.nan]
        if len(ids_closed) > 0:
            ids_closed = np.array(ids_closed)
            j_end2[ids_closed] = -2  # avoid searching these
            j_start2[ids_closed] = -2
        xy_start = np.array(xy_start)
        xy_end = np.array(xy_end)

        for s in range(nSub):
            # end of current to start of other?
            id_join, reverse = join_to(s, j_end2, j_start2, xy_end, xy_start)
            if id_join is not None:
                j_end2[s] = id_join
                if reverse:
                    j_end2[id_join] = s
                    j_end2end[s] = True
                    j_end2end[id_join] = True
                else:
                    j_start2[id_join] = s

            # start of current to end of other?
            id_join, reverse = join_to(s, j_start2, j_end2, xy_start, xy_end)
            if id_join is not None:
                j_start2[s] = id_join
                if reverse:
                    j_start2[id_join] = s
                    j_start2start[s] = True
                    j_start2start[id_join] = True
                else:
                    j_end2[id_join] = s

        self.subpaths_dict["join_start2"] = j_start2
        self.subpaths_dict["join_end2"] = j_end2
        self.subpaths_dict["join_start2start"] = j_start2start
        self.subpaths_dict["join_end2end"] = j_end2end

    def join_sub(self, sub1, sub2):
        """Join Path() by interpolation or straight line segment."""
        if self.options.option_join == "1":
            # generate new interpolated join node from end/start nodes
            x2, y2, x3, y3, x_end, y_end = sub1[-1].args
            x_start, y_start = sub2[0].args
            join_x, join_y = 0.5 * (x_end + x_start), 0.5 * (y_end + y_start)
            # remove end/start + input new joined node
            sub1[-1] = Curve(x2, y2, x3, y3, join_x, join_y)
            sub2.pop(0)  # pop M
        elif self.options.option_join == "2":
            # straight line
            x, y = sub2[0].args
            sub2[0] = Line(x, y)

        new_sub = sub1 + sub2

        return new_sub

    def join_subpaths(self):
        """Join different subpaths where end nodes closer than threshold."""

        def join_sub(id_this, end=True):
            # join one by one until -1 or back to s (closed)
            close_this = False
            sub_this = self.subpaths_dict["subpaths"][id_this]
            if end:
                id_join = self.subpaths_dict["join_end2"][id_this]
                reverse = self.subpaths_dict["join_end2end"][id_this]
            else:
                id_join = self.subpaths_dict["join_start2"][id_this]
                reverse = self.subpaths_dict["join_start2start"][id_this]
            sub_join = self.subpaths_dict["subpaths"][id_join]
            if reverse:
                sub_join = sub_join.reverse()
            if end:
                joined_sub = self.join_sub(sub_this, sub_join)
            else:
                joined_sub = self.join_sub(sub_join, sub_this)
            moved_to[id_join] = id_this
            prev = id_this
            # continue if id_join joined to more
            if (
                self.subpaths_dict["join_end2"][id_join] > -1
                and self.subpaths_dict["join_start2"][id_join] > -1
            ):
                # already joined so both joined if continue
                proceed = True
                while proceed:
                    if self.subpaths_dict["join_end2"][id_join] != prev:
                        id_next = self.subpaths_dict["join_end2"][id_join]
                        reverse_join = self.subpaths_dict["join_end2end"][id_join]
                    else:
                        id_next = self.subpaths_dict["join_start2"][id_join]
                        reverse_join = self.subpaths_dict["join_start2start"][id_join]
                    if reverse_join:
                        reverse = not reverse
                    if moved_to[id_next] == id_this:
                        close_this = True
                        proceed = False
                    else:
                        sub_join = self.subpaths_dict["subpaths"][id_next]
                        if reverse:
                            sub_join = sub_join.reverse()
                        if end:
                            joined_sub = self.join_sub(joined_sub, sub_join)
                        else:
                            joined_sub = self.join_sub(sub_join, joined_sub)
                        moved_to[id_next] = id_this
                        if (
                            self.subpaths_dict["join_end2"][id_next] > -1
                            and self.subpaths_dict["join_start2"][id_next] > -1
                        ):
                            prev = id_join
                            id_join = id_next
                        else:
                            proceed = False
            self.subpaths_dict["subpaths"][s] = joined_sub

            return close_this

        if self.subpaths_dict["open"].count(True) > 1:  # any subpaths to join
            # fill logics arrays
            self.find_join_matches()

            # fill new_path according to logics
            n_sub = len(self.subpaths_dict["subpaths"])
            moved_to = np.arange(n_sub)
            for s in range(n_sub):
                if moved_to[s] == s:  # not joined yet
                    close_this = False
                    if self.subpaths_dict["join_end2"][s] > -1:
                        close_this = join_sub(s, end=True)

                    if (
                        self.subpaths_dict["join_start2"][s] > -1
                        and close_this is False
                    ):
                        close_this = join_sub(s, end=False)

                    # close the new subpath if start/end node is closer than maxdist
                    if close_this:
                        this_sub = self.subpaths_dict["subpaths"][s]
                        if self.options.option_join == "1":
                            x_start, y_start = this_sub[0].args
                            x2, y2, x3, y3, x_end, y_end = this_sub[-1].args
                            join_x = 0.5 * (x_start + x_end)
                            join_y = 0.5 * (y_start + y_end)
                            this_sub[0] = Move(join_x, join_y)
                            this_sub[-1] = Curve(x2, y2, x3, y3, join_x, join_y)
                        this_sub.append(ZoneClose())

        # remove all subpaths that are joined to others
        remove_ids = []
        for subno, m_to in enumerate(moved_to):
            if m_to != subno:
                remove_ids.append(subno)
        remove_ids.sort(reverse=True)
        for rem_no in remove_ids:
            self.subpaths_dict["subpaths"].pop(rem_no)


if __name__ == "__main__":
    CleanUpPath().run()



#!/usr/bin/env python3

import sys

root = {}
children = {}

for line in sys.stdin:
    line = (
        line.rstrip()
        .replace("|", " ")
        .replace("--", "  ")
        .replace("`", " ")
        .replace("└", " ")
        .replace("├", " ")
        .replace("│", " ")
        .replace("─", " ")
    )

    record = line.strip()
    pkg, version, *_ = record.split(" ", 2)
    tab = len(line) - len(record)

    if tab == 0:
        root[pkg] = version
    else:
        children[pkg] = version

print("# Packages Used Directly\n\nYou must always include these in your package:\n")
for pkg, version in root.items():
    print(f" * {pkg} {version}")

print("\n")
print(
    "# Packages Used Indirectly\n\nYou may or may not need to include these, depending on if your packaging system automatically pulls in sub-deps:\n"
)
for pkg, version in children.items():
    print(f" * {pkg} {version}")
